/*
Copyright (c) 2007 Health Market Science, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

You can contact Health Market Science at info@healthmarketscience.com
or at the following address:

Health Market Science
2700 Horizon Drive
Suite 200
King of Prussia, PA 19406
*/

package com.healthmarketscience.common.util;

import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Static utility methods for working with classes and reflection.
 *
 */
public class ClassUtil {

  /**
   * Get a collection of fields of the given object with the given annotation.
   * 
   * @param <AType> the type of the annotation to look for.
   * @param instance the object to scan for annotated fields.
   * @param annotationClass the type of the annotation to look for.
   * @param inherit whether to get fields from super classes.
   * @return a collection of pairs of fields and their annotations
   */
  public static <AType extends Annotation> 
  Collection<Tuple2<AType,Field>> findAnnotatedFields(
          Object instance, 
          final Class<AType> annotationClass, 
          boolean inherit) {
    return findAnnotatedElements(annotationClass, 
            getFields(instance.getClass(), inherit));
  }
  
  /**
   * Get a collection of methods of the given object with the given annotation.
   * 
   * @param <AType> the type of the annotation to look for.
   * @param instance the object to scan for annotated methods.
   * @param annotationClass the type of the annotation to look for.
   * @param inherit whether to get methods  from super classes.
   * @return a collection of pairs of methods and their annotations
   */
  public static <AType extends Annotation> 
  Collection<Tuple2<AType,Method>> findAnnotatedMethods(
          Object instance, 
          final Class<AType> annotationClass, 
          boolean inherit) {
    return findAnnotatedElements(annotationClass, 
            getMethods(instance.getClass(), inherit));
  }
  
  

  /**
   * Search for elements (fields or methods) which have the given
   * annotation.
   * 
   * @param annotationClass the annotation to search for
   * @param members the members to search over
   * @return a collection of the members with the annotation (an empty 
   *         collection if none are found)
   */
  public static <AType extends Annotation, MType extends AnnotatedElement> 
  Collection<Tuple2<AType,MType>> findAnnotatedElements( 
          Class<AType> annotationClass, MType... members ) {
    List<Tuple2<AType,MType>> res = new ArrayList<Tuple2<AType,MType>>();
    for (MType member : members) {
      AType ann = member.getAnnotation(annotationClass);
      if (null != ann) {
        res.add(Tuple2.create(ann, member));
      }
    }
    return res;
  }

  /**
   * Test if the class is Boolean or boolean.
   */
  public static boolean isBoolean( Class<?> clazz ) {
    return clazz.equals(boolean.class) 
        || clazz.equals(Boolean.class);
  }
  
  /**
   * Returns whether or not the given member should be visible to the given
   * class.
   *  
   * @param clazz the class that wants to see the given member
   * @param member the member
   * @return whether or not the given member should be visible to the given 
   *         class
   */
  private static boolean shouldSee(Class<?> clazz, Member member) {
    
    int modifiers = member.getModifiers();

    // It's visible if it's declared in this class or if it's protected
    if (member.getDeclaringClass() == clazz ||
        Modifier.isProtected(modifiers)) {
      return true;
    }    
    
    // If it wasn't declared in this class, it's invisible if it's private
    if (Modifier.isPrivate(modifiers)) {
      return false;
    }
    
    // Otherwise it's visible if we're in the same package
    return member.getDeclaringClass().getPackage().equals(clazz.getPackage());
  }

  private static Field[] getFields(Class<?> clazz, boolean inherit) {
    Set<Field> fields = getFields(clazz, clazz, inherit);
    return fields.toArray(new Field[fields.size()]);
  }
  
  private static Method[] getMethods(Class<?> clazz, boolean inherit) {
    Set<Method> methods = getMethods(clazz, clazz, inherit);
    return methods.toArray(new Method[methods.size()]);
  }

  
  private static Set<Field> getFields(Class<?> clazz, Class<?> lowestClass, boolean inherit) {
    
    // Use linked hash set to preserve insertion order
    Set<Field> fields = new LinkedHashSet<Field>();
    
    for (Field field : clazz.getDeclaredFields()) {
      if (shouldSee(lowestClass, field)) {
        fields.add(field);
      }
    }
    Class<?> superClass = clazz.getSuperclass();
    if (superClass != null && inherit) {
      fields.addAll(getFields(superClass, clazz, true));
    }
    return fields;
  }
  
  
  private static Set<Method> getMethods(
          Class<?> clazz, 
          Class<?> lowestClass,
          boolean inherit) {
    
    // Use linked hash set to preserve insertion order
    Set<Method> methods = new LinkedHashSet<Method>();
    
    for (Method method : clazz.getDeclaredMethods()) {
      if (shouldSee(lowestClass, method)) {
        methods.add(method);        
      }
    }
    
    Class<?> superClass = clazz.getSuperclass();
    if (superClass != null && inherit) {
      methods.addAll(getMethods(superClass, clazz, true));
    }
    return methods;
  }
  
}
